#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <stdint.h>
#include <errno.h>
#include <arpa/inet.h>

#include "galois.h"
#include "gmath.h"
#include "cJSON.h"
#include "list.h"

struct fragment
 {	struct fragment	*prev,*next;
	int		f_position;
	int		f_rs_length;
	uint32_t	f_rs_key;
	unsigned char	*f_data;
 };	

struct inode
 {	struct inode	*prev,*next;
	int		in_inode;
	char		*in_filename;
	int		in_size;
	int		in_offset;
	struct fragment	*in_first_f,*in_last_f;
	int		in_fragment_length;
	int		in_fragment_count;
 };

char *freadline(FILE *fr)
{
 char   *ret,buff[256];

 ret=NULL;

 if ( feof(fr) )
	return(NULL);

 while ( 1 )
  {	if ( fgets(buff,255,fr)==NULL ) 
		break;

	if ( ret==NULL )
		ret=(char *)malloc(strlen(buff)+1),ret[0]=0;
	else	
		ret=(char *)realloc(ret,strlen(ret)+strlen(buff)+1);

	strcat(ret,buff);
	if ( ret[strlen(ret)-1]==10 )   
		break;
  }

 return(ret);
}


int fprint_gmatrix_row_cols(FILE *fw,uint32_t **m,int nrow,int ncol)
{
 int	i;
 for ( i=0; i<nrow; i++ )
  {	int	j;
	for ( j=0; j<ncol; j++ )
	 {	fprintf(fw,"%10u ",m[i][j]);
	 }
	fprintf(fw,"\n");
  }
 return(0);
}

int fprint_gmatrix(FILE *fw,uint32_t **m,int n)
{
 fprint_gmatrix_row_cols(fw,m,n,n);
 return(0);
}



int fprint_usage(FILE *fw,char *argv0)
{
 fprintf(fw,"Usage:\t%s [-h|--help] [-d|--directory <dir>] [-v|--verbose] <input.json>\n",argv0);
 return(0);
}

int do_recovery(char *output_directory,struct inode *in)
{
 int			in_fragment_needed;
 uint32_t		**rs_matrix,**rs_echelon;
 struct fragment	*f;
 int			row,fragment_length;
 int			*pivot_index,i,w;
 struct fragment	**list_f;
 char			out_filename[256];

 uint32_t		*out_fragment;

 FILE			*fw;

 fragment_length=in->in_fragment_length;
 
 in_fragment_needed=(in->in_size+fragment_length-1)/fragment_length;
 if ( in->in_fragment_count < in_fragment_needed )
  {	return(in_fragment_needed-in->in_fragment_count);
  }

 out_fragment=malloc(fragment_length);
 pivot_index=malloc(sizeof(int)*in_fragment_needed);
 list_f=malloc(sizeof(struct fragment *)*in->in_fragment_count);

 rs_echelon=matrix_alloc_gen(in_fragment_needed,in->in_fragment_count,sizeof(uint32_t));
 rs_matrix=matrix_alloc_gen(in->in_fragment_count,in_fragment_needed,sizeof(uint32_t));

 for ( i=0; i<in_fragment_needed; i++ )
  {	memset(rs_echelon[i],0,sizeof(uint32_t)*in->in_fragment_count);
  }

 for ( f=in->in_first_f,row=0; f != NULL; f=f->next, row++ )
  {	int		index,j;
	uint32_t	t;
	memset(rs_matrix [row],0,sizeof(uint32_t)*in_fragment_needed);
	index=(f->f_position-in->in_offset)/fragment_length;
	if ( ! ( 0<=index && index<in_fragment_needed ) )
	 {	fprintf(stderr,"index=%d (in_fragment_needed=%d)\n",index,in_fragment_needed);
		exit(2);
	 }
	t=1;
	rs_matrix[row][index]=0;
	rs_echelon[index][row]=0;
	for ( j=0; j<f->f_rs_length; j++ )
	 {	rs_matrix[row][index]^=t;
		rs_echelon[index][row]^=t;
		t=galois_2p32_mul(t,f->f_rs_key);
		index++;
		if ( index==in_fragment_needed )
			index=0;
	 }
  }

/*
 fprintf(stderr,"[1] needed=%d in_fragment_count=%d\n",in_fragment_needed,in->in_fragment_count);
*/

 galois_2p32_row_echelon(rs_echelon,in_fragment_needed,in->in_fragment_count);
/* fprint_gmatrix_row_cols(stderr,rs_echelon,in->in_fragment_count,in_fragment_needed); */

/*
 fprintf(stderr,"[2]\n");
*/

 for ( i=0,w=0; i<in_fragment_needed; i++ )
  {	while ( w<in->in_fragment_count )
	 {	if ( rs_echelon[i][w] != 0 )
			break;
		w++;
	 }
	if ( w<in->in_fragment_count )
	 {	pivot_index[i]=w;
	 }
	else
	 {	matrix_free(rs_matrix);
		matrix_free(rs_echelon);
		free(list_f);
		free(pivot_index);
		free(out_fragment);
		/*
		fprintf(stderr,"galois_2p32_column_echelon(): had=%d needed=%d\n",i,in_fragment_needed);
		*/
		return(in_fragment_needed-i);
	 }
	w++; 
  }



 for ( i=0; i<in_fragment_needed; i++ )
  {	int	j;
	j=pivot_index[i];
	if ( i<j )
		memcpy(rs_matrix[i],rs_matrix[j],sizeof(uint32_t)*in_fragment_needed);
  }

 for ( f=in->in_first_f,row=0; f != NULL ; f=f->next, row++ )
  {	list_f[row]=f;
  }

 galois_2p32_invert_gauss(rs_matrix,in_fragment_needed); 

 if ( 0<in->in_offset && strrchr(in->in_filename,'.')==NULL )
  {	snprintf(out_filename,256,"%s-%d",in->in_filename,in->in_offset);
  }
 else if ( 0<in->in_offset )
  {	char	*p;
	int	base_len;
	p=strrchr(in->in_filename,'.');
	base_len=p-in->in_filename;
	memcpy(out_filename,in->in_filename,base_len);
	out_filename[base_len]=0;
	sprintf(out_filename+base_len,"_%d%s",in->in_offset,p);
  }
 else
  {	snprintf(out_filename,256,"%s",in->in_filename);
  }

 fw=fopen(out_filename,"w");

 for ( f=in->in_first_f,row=0; f != NULL && row<in_fragment_needed; f=f->next, row++ )
  {	uint32_t	*grow;
	int		col;

	grow=rs_matrix[row];
	memset(out_fragment,0,fragment_length);

	for ( col=0; col<in_fragment_needed; col++ )
	 {	uint32_t	*data;
		int		i;
		data=(uint32_t *)(list_f[pivot_index[col]]->f_data);
		for ( i=0; i<fragment_length/4; i++ )
		 {	uint32_t	t;
			t=galois_2p32_mul(grow[col],ntohl(data[i]));
			out_fragment[i] ^= htonl(t);
		 }
	 }
	if ( row<in_fragment_needed-1 )	
		fwrite(out_fragment,1,fragment_length,fw);
	else
	 {	int	last_fragment_size;
		last_fragment_size=fragment_length-(fragment_length*in_fragment_needed-in->in_size);
		fwrite(out_fragment,1,last_fragment_size,fw);
	 }

  }

 fclose(fw);

 matrix_free(rs_matrix);
 matrix_free(rs_echelon);
 free(list_f);
 free(pivot_index);
 free(out_fragment);

 return(0);
}

int main(int argc,char *argv[])
{
 FILE		*fr;
 int		i;
 char		*input_json_file,*output_directory;
 int		lcount;
 int		is_verbose;

 struct inode	*first_in,*last_in;

 input_json_file=NULL;
 output_directory=".";

 is_verbose=0;

 for ( i=1; i<argc; i++ )
  {	if ( strcmp(argv[i],"-h")==0 || strcmp(argv[i],"--help")==0 )
	 {	fprint_usage(stdout,argv[0]);
		return(0);
	 }
	else if ( ( strcmp(argv[i],"-d")==0 || strcmp(argv[i],"--directory")==0 ) && i<argc-1 )
	 {	i++;
		output_directory=argv[i];
	 }
	else if ( strcmp(argv[i],"-v")==0 || strcmp(argv[i],"--verbose")==0 )
		is_verbose++;
	else if ( argv[i][0]=='-' )
	 {	fprintf(stderr,"%s: error: invalid command line argument near '%s'.\n",argv[0],argv[i]);
		return(1);
	 }
	else
		input_json_file=argv[i];
  }

 if ( input_json_file==NULL || strcmp(input_json_file,"-")==0 )
	fr=stdin;
 else if ( (fr=fopen(input_json_file,"r"))==NULL )
  {	fprintf(stderr,"%s: error: unable to open input file '%s': %s.\n",argv[0],input_json_file,strerror(errno));
	return(1);
  }
 
 first_in=NULL;
 last_in=NULL;

 lcount=0;
 while ( ! feof(fr) )
  {	char		*line;
	cJSON		*json,*j;

	char		*filename;
	char		*fragment;
	int		size,inode,offset;
	uint32_t	position;
	int		rs_length;
	int		rs_key;
	int		length;

	struct inode	*in;
	struct fragment	*f;

	int		i;

	line=freadline(fr);

	if ( line==NULL )
		break;

	lcount++;

	json = cJSON_Parse(line);
	if ( json==NULL )
	 {	fprintf(stderr,"%s: warning: unable to parse input line %d as JSON, skipped.\n",argv[0],lcount);
		free(line);
		continue;
	 }
	else if ( json->type != cJSON_Object )
	 {	fprintf(stderr,"%s: warning: input line %d is not a JSON object.\n",argv[0],lcount);
		cJSON_Delete(json);	
		free(line);
		continue;
	 }

	filename=NULL;
	size=0;
	offset=0;
	inode=0;
	position=0;
	rs_length=1;
	rs_key=0;
	length=0;
	fragment=NULL;

	for ( j=json->child; j != NULL; j=j->next )
	 {	/* fprintf(stderr,"[%s] [%d]\n",j->string,j->type); */
		if ( strcmp(j->string,"filename")==0 && j->type==cJSON_String )
			filename=j->valuestring;
		else if ( strcmp(j->string,"inode")==0 && j->type==cJSON_Number )
			inode=(int)j->valuedouble;
		else if ( strcmp(j->string,"size")==0 && j->type==cJSON_Number )
			size=(int)j->valuedouble;
		else if ( strcmp(j->string,"offset")==0 && j->type==cJSON_Number )
			offset=(int)j->valuedouble;
		else if ( strcmp(j->string,"position")==0 && j->type==cJSON_Number )
			position=(int)j->valuedouble;
		else if ( strcmp(j->string,"rs_length")==0 && j->type==cJSON_Number )
			rs_length=(int)j->valuedouble;
		else if ( strcmp(j->string,"rs_key")==0 && j->type==cJSON_Number )
			rs_key=(int)j->valuedouble;
		else if ( strcmp(j->string,"length")==0 && j->type==cJSON_Number )
			length=(int)j->valuedouble;
		else if ( strcmp(j->string,"fragment")==0 && j->type==cJSON_String )
			fragment=j->valuestring;
	 }

	/*
	fprintf(stderr,"%s: size=%d position=%d length=%d\n",filename,size,position,length);
	*/

	for ( in=first_in; in != NULL ; in=in->next )
	 {	if ( in->in_inode==inode )
			break;
	 }
	if ( in==NULL )
	 {	in=dlist_new(struct inode);
		dlist_insert_last(first_in,last_in,in);
		in->in_inode=inode;
		in->in_first_f=NULL;
		in->in_last_f=NULL;
		in->in_size=0;
		in->in_offset=0;
		in->in_filename=NULL;
		in->in_fragment_length=0;
		in->in_fragment_count=0;
	 }
	if ( in->in_filename==NULL && filename != NULL )
	 {	in->in_filename=strdup(filename);
		in->in_size=size;
		in->in_offset=offset;
	 }

	if ( 0<in->in_fragment_length && in->in_fragment_length != length )
	 {	fprintf(stderr,"%s: warning: variable fragment length sizes are not supported at the present moment (line %d), fragment skipped.\n",argv[0],lcount);
		cJSON_Delete(json);	
		free(line);
		continue;
	 }
	else if ( fragment==NULL || strlen(fragment) != 2*length )
	 {	fprintf(stderr,"%s: warning: fragment is missing and/or fragment data block has unexpected length (line %d), skipped.\n",argv[0],lcount);
		cJSON_Delete(json);	
		free(line);
		continue;
	 }
	else if ( in->in_fragment_length <= 0 )
		in->in_fragment_length=length;

	f=dlist_new(struct fragment);
	dlist_insert_last(in->in_first_f,in->in_last_f,f);
	f->f_position=position;
	f->f_rs_key=rs_key;
	f->f_rs_length=rs_length;
	f->f_data=malloc(length);
	for ( i=0; i<length; i++ )
	 {	char		hex[8];
		unsigned	databyte;
		hex[0]=fragment[2*i+0];
		hex[1]=fragment[2*i+1];
		hex[2]=0;
		sscanf(hex,"%x",&databyte);
		f->f_data[i]=databyte;
	 }
	in->in_fragment_count++;

	cJSON_Delete(json);
	free(line);
  }

 fclose(fr);

 if ( 1 )
  {	struct inode	*in;
	for ( in=first_in; in != NULL; in=in->next )
	 {	int	in_fragment_needed,in_fragment_missing;
		if ( in->in_size <= 0 || in->in_filename==NULL )
	 	 {	fprintf(stderr,"%s: warning: inode %d has no associated filesystem data (size and name).\n",argv[0],in->in_inode);
			continue;
		 }
		in_fragment_needed=(in->in_size+in->in_fragment_length-1)/in->in_fragment_length;
		in_fragment_missing=do_recovery(output_directory,in);
		if ( 1<=is_verbose && 0<in_fragment_missing )
		 {	fprintf(stderr,"%s: warning: %s: missing %d/%d fragments.\n",argv[0],in->in_filename,in_fragment_missing,in_fragment_needed);
		 }
		if ( 2<=is_verbose && ! in_fragment_missing )
		 {	fprintf(stderr,"%s: note: %s: done (offset: %d).\n",argv[0],in->in_filename,in->in_offset);
		 }
	 }
  }

/*
  m=galois_2p32_matrix_alloc(4);
 for ( i=0; i<4; i++ )
  {	for ( j=0; j<4; j++ )
	 {	m[i][j]=0;
	 }
	m[i][i]=1;
  }
 m[0][1]=42;
 m[1][0]=137;


 fprint_gmatrix(stdout,m,4);
 fprintf(stdout,"\n");

 galois_2p32_invert_gauss(m,4);
 fprint_gmatrix(stdout,m,4);
 fprintf(stdout,"\n");

 galois_2p32_invert_gauss(m,4);
 fprint_gmatrix(stdout,m,4);
 fprintf(stdout,"\n");
*/

 return(0);
}
